
##### global variables

replicates = 100
recalculate.lk = 5 # 10
attempts.max = 5 # 5
crown = TRUE
df1 = 0.5 # m1 vs m2
df2 = 2 # m1 vs m3
degrees.of.freedom = 1.5 # m2 vs m3
debugging = FALSE
print.rv.file = FALSE

### functions

randomRows<-function(df,n){
  return(df[sample(nrow(df),n),])
}

simulate.data <- function(tr, frs, rho, log.file = ""){
  
  ## estimate lambda and mu from empirical data
  convergence = 1
  while(convergence != 0){
    est <- fbdR::est.bd.extant(1, 0.5, tr, crown = T, rho = rho)
    lambda = est$par[[1]]
    mu = est$par[[2]]
    convergence = est$convergence
  }
  lambda.branching <<- lambda
  mu.branching <<- mu
  
  ## if turnover > 0.9 it becomes challenging to simulate large trees (e.g. with lots of tips that survive to the present)
  if((mu/lambda) > 0.9)
    mu = .9 * lambda
  
  ## simulate data under the null (constrained) model
  trees = list()
  fossils = list()
  ls = c()
  ms = c()
  root.ages = c()
  
  tips = ceiling( length(tr$tip.label) / rho)
  
  target.range.length = length(frs$clade)
  
  for(i in 1:reps){
    
    ## simulate trees
    
    l = lambda
    m = mu
    
    #l = 0
    #m = 1
    # while(l < m){
    #   l = rexp(1,1/lambda)
    #   m = rexp(1,1/mu)
    # }
    
    trees[[i]] <- TreeSim::sim.bd.taxa(tips,1,l,m)[[1]]
    
    ages = TreeSim::getx(trees[[i]],sersampling = 1)
    root.age = max(as.data.frame(ages)$times)
    root.ages = c(root.ages, root.age)
    
    #print(i)
    #print(root.age)
    
    ## simulate fossils
    f = 0.5
    la = 0
    
    current.range.number = 0
    
    while(current.range.number < target.range.length){
      
      la = la + 0.1
      
      # add symmetric species 
      ages1 <- FossilSim::mixed.ages(trees[[i]], f)
      # add anagenic species 
      ages2 <- FossilSim::anagenic.species(ages1, la)
      
      current.range.number = length(ages2$sp)
      
    }
    
    fossils[[i]] <- ages2
    
    ## book keeping
    ls = c(ls, (2*l) + la - (l*f))
    ms = c(ms, (m + l) + la - (l*f))
    
  }
  
  if(print.rv.file)
    write.table(data.frame(ls,ms), row.names = F, quote = F, file = rv.file)
  
  cat("lambda.star =", mean(ls), "\n", file = log.file, append = T)
  cat("mu.star =", mean(ms), "\n", file = log.file, append = T)
  cat("Root.ages =", mean(root.ages), "\n", file = log.file, append = T)
  return(list(trees=trees,fossils=fossils,root.ages=root.ages))
  #eof 
}

calculate.dAIC <- function(data, rho, ranges, log.file = ""){
  fossils<-data$fossils
  trees<-data$trees
  
  dAIC = c()
  
  for(i in 1:length(fossils)){
    cat("rep ", i, "\n", sep="", file = log.file, append = T)
    
    tree1 <- trees[[i]]
    if(rho < 1){
      tree1 <- geiger::drop.extinct(tree1)
      n = ceiling(length(tree1$tip.label) * (1-rho))
      tree1 <- geiger::drop.random(tree1, n)
    }
    
    frs <- fossils[[i]]
    frs = randomRows(frs, ranges)
    
    # contrained
    lnL = -10^100
    for(j in 1:recalculate.lk){
      attempts = 0
      convergence = 1
      while(convergence != 0){
        b = runif(1)
        b.star = runif(1, b, 2)
        nd = runif(1, 0, b*.99)
        est <- fbdR::est.bd.combined(tree1,frs,b=b,b.star=b.star,nd=nd,crown=TRUE,rho=rho,constrained=TRUE)
        convergence = est$convergence
        if(attempts == 100)
          break
        attempts = attempts + 1
      }
      if(convergence != 0){
        next
      } else if(est$value > lnL) {
        lnL = est$value
        # note we do not need to store the parameter estimates
      }
      cat("j ", j, " lnL null = ", lnL, "\n", sep="", file = log.file, append = T)
    }
    if(convergence != 0)
      next
    AIC.null = -(2 * lnL) + (2 * 2.5) # should this be 3 or 2.5?
    
    # unconstrained
    lnL = -10^100
    for(j in 1:recalculate.lk){
      attempts = 0
      convergence = 1
      while(convergence != 0){
        b = runif(1)
        d = runif(1, 0, b*.99)
        b.star = runif(1, b, 2)
        d.star = runif(1, 0, b.star*.99)
        est <- fbdR::est.bd.combined(tree1,frs,b=b,d=d,b.star=b.star,d.star=d.star,crown=TRUE,rho=rho,constrained=FALSE)
        convergence = est$convergence
        if(attempts == 100)
          break
        attempts = attempts + 1
      }
      if(convergence != 0){
        next
      } else if(est$value > lnL) {
        lnL = est$value
      }
      cat("j ", j, " lnL alt = ", lnL, "\n", sep="", file = log.file, append = T)
    }
    if(convergence != 0)
      next
    AIC.alt = -(2 * lnL) + (2 * 4)
    
    cat("dAIC ", AIC.null - AIC.alt, "\n", sep="", file = log.file, append = T)
    
    # calculate delta AIC
    dAIC = c(dAIC, AIC.null - AIC.alt)
  }
  return(list(dAIC=dAIC))
  #eof
}

calculate.LR <- function(data, rho, ranges, log.file = ""){
  fossils<-data$fossils
  trees<-data$trees
  
  LRs = c()
  
  for(i in 1:length(fossils)){
    cat("rep ", i, "\n", sep="", file = log.file, append = T)
    
    tree1 <- trees[[i]]
    if(rho < 1){
      tree1 <- geiger::drop.extinct(tree1)
      n = ceiling(length(tree1$tip.label) * (1-rho))
      tree1 <- geiger::drop.random(tree1, n)
    }
    
    frs <- fossils[[i]]
    frs = randomRows(frs, ranges)
    
    # contrained
    lnL = -10^100
    for(j in 1:recalculate.lk){
      attempts = 0
      convergence = 1
      while(convergence != 0){
        b = runif(1)
        b.star = runif(1, b, 2)
        nd = runif(1, 0, b*.99)
        est <- fbdR::est.bd.combined(tree1,frs,b=b,b.star=b.star,nd=nd,crown=TRUE,rho=rho,constrained=TRUE)
        convergence = est$convergence
        if(attempts == 100)
          break
        attempts = attempts + 1
      }
      if(convergence != 0){
        next
      } else if(est$value > lnL) {
        lnL = est$value
      }
      cat("j ", j, " lnL null = ", lnL, "\n", sep="", file = log.file, append = T)
    }
    if(convergence != 0)
      next
    lk.null = lnL
    
    # unconstrained
    lnL = -10^100
    for(j in 1:recalculate.lk){
      attempts = 0
      convergence = 1
      while(convergence != 0){
        b = runif(1)
        d = runif(1, 0, b*.99)
        b.star = runif(1, b, 2)
        d.star = runif(1, 0, b.star*.99)
        est <- fbdR::est.bd.combined(tree1,frs,b=b,d=d,b.star=b.star,d.star=d.star,crown=TRUE,rho=rho,constrained=FALSE)
        convergence = est$convergence
        if(attempts == 100)
          break
        attempts = attempts + 1
      }
      if(convergence != 0){
        next
      }
      if(est$value > lnL){
        lnL = est$value
      }
      cat("j ", j, " lnL alt = ", lnL, "\n", sep="", file = log.file, append = T)
    }
    if(convergence != 0)
      next
    lk.alt = lnL
    
    # calculate lk ratio
    LR = 2 * (lk.alt - lk.null)
    
    cat("LR ", LR, "\n", sep="", file = log.file, append = T)
    
    LRs = c(LRs, LR)
  }
  return(list(LRs=LRs))
  #eof
}

likelihood.ratio.tests<-function(trees, fossils, rho, ranges, file = "", shifts = FALSE, file2 = "log.txt"){
  
  ## parameters
  
  ## data to collect
  
  m1.birth.extant = c()
  m1.death.extant = c()
  m1.birth.ranges = c()
  m1.death.ranges = c()
  m1.lk = c()
  m1.convergence = c()
  
  m2.birth.extant = c()
  m2.death.extant = c()
  m2.birth.ranges = c()
  m2.death.ranges = c()
  m2.lk = c()
  m2.convergence = c()
  
  m3.birth.extant = c()
  m3.death.extant = c()
  m3.birth.ranges = c()
  m3.death.ranges = c()
  m3.lk = c()
  m3.convergence = c()
  
  for(i in 1:replicates){
    
    tree1<-trees[[i]]
    if(rho < 1){
      tree1 <- geiger::drop.extinct(tree1)
      n = ceiling(length(tree1$tip.label) * (1-rho))
      tree1 <- geiger::drop.random(tree1, n)
    }
    
    trees[[i]]<-tree1
    
    frs<-fossils[[i]]
    frs = randomRows(frs, ranges)
    
    fossils[[i]]<-frs
    
  }
  
  if(debugging)
    cat("Model 1...\n")
  
  ## model 1
  ## lambda = lambda.star and mu = mu.star
  for(i in 1:replicates){
    
    if(debugging)
      cat(i,"\n")
    
    tree1<-trees[[i]]
    frs<-fossils[[i]]
    
    lnL = -10^100
    tmp = NULL
    
    for(j in 1:recalculate.lk){
      
      attempts = 1
      convergence = 1
      
      while(convergence != 0){
        
        b = runif(1)
        d = runif(1, 0, b*.99)
        
        est = fbdR::est.bd.combined(tree1, frs, b = b, d = d, constrained = TRUE, crown = crown, rho = rho, constrained.p = 2)
        
        convergence = est$convergence
        
        if(attempts == attempts.max)
          break
        
        attempts = attempts + 1
      }
      
      if(convergence != 0)
        next
      else{
        if(est$value > lnL){
          lnL = est$value
          tmp = est
        }
      }
    }
    
    # lk estimation has not converged
    if(is.null(tmp)){
      m1.birth.extant = c(m1.birth.extant, NA)
      m1.death.extant = c(m1.death.extant, NA)
      m1.birth.ranges = c(m1.birth.ranges, NA)
      m1.death.ranges = c(m1.death.ranges, NA)
      m1.lk = c(m1.lk, NA)
      m1.convergence = c(m1.convergence, est$convergence)
    } else {
      est = tmp
      m1.birth.extant = c(m1.birth.extant, est$par[1])
      m1.death.extant = c(m1.death.extant, est$par[2])
      m1.birth.ranges = c(m1.birth.ranges, est$par[1])
      m1.death.ranges = c(m1.death.ranges, est$par[2])
      m1.lk = c(m1.lk, est$value)
      m1.convergence = c(m1.convergence, est$convergence)
    }
  }
  
  cat("finished M1...\n", file=file2, append=T)
  
  if(debugging)
    cat("Model 2...\n")
  
  ## model 2
  ## constrained model with lambda, lambda.star and d (= d*)
  for(i in 1:replicates){
    
    if(debugging)
      cat(i,"\n")
    
    tree1<-trees[[i]]
    frs<-fossils[[i]]
    
    lnL = -10^100
    tmp = NULL
    
    for(j in 1:recalculate.lk){
      
      attempts = 1
      convergence = 1
      
      while(convergence != 0){
        
        b = runif(1)
        b.star = runif(1, b, 2)
        nd = runif(1, 0, b*.99)
        
        est = fbdR::est.bd.combined(tree1, frs, b = b, b.star = b.star, nd = nd, constrained = TRUE, crown = crown, rho = rho)
        
        convergence = est$convergence
        
        if(attempts == attempts.max)
          break
        
        attempts = attempts + 1
      }
      
      if(convergence != 0)
        next
      else{
        if(est$value > lnL){
          lnL = est$value
          tmp = est
        }
      }
    }
    
    # lk estimation has not converged
    if(is.null(tmp)){
      m2.birth.extant = c(m2.birth.extant, NA)
      m2.death.extant = c(m2.death.extant, NA)
      m2.birth.ranges = c(m2.birth.ranges, NA)
      m2.death.ranges = c(m2.death.ranges, NA)
      m2.lk = c(m2.lk, NA)
      m2.convergence = c(m2.convergence, est$convergence)
    } else {
      est = tmp
      m2.birth.extant = c(m2.birth.extant, est$par[1])
      m2.birth.ranges = c(m2.birth.ranges, est$par[2])
      m2.death.extant = c(m2.death.extant, est$par[1]-est$par[3])
      m2.death.ranges = c(m2.death.ranges, est$par[2]-est$par[3])
      m2.lk = c(m2.lk, est$value)
      m2.convergence = c(m2.convergence, est$convergence)
    }
  }
  
  cat("finished M2...\n", file=file2, append=T)
  
  if(debugging)
    cat("Model 3...\n")
  
  ## model 3
  ## lambda = lambda.star and mu = mu.star 
  for(i in 1:replicates){
    
    if(debugging)
      cat(i,"\n")
    
    #if(debugging)
    #cat("lambda\t", "mu\t", "lambda*\t", "mu*\t", "lk\n", file="debugging.txt", append=F)
    
    tree1<-trees[[i]]
    frs<-fossils[[i]]
    
    lnL = -10^100
    tmp = NULL
    
    for(j in 1:recalculate.lk){
      
      attempts = 1
      convergence = 1
      
      while(convergence != 0){
        
        b = runif(1)
        d = runif(1, 0, b*.99)
        b.star = runif(1, b, 2)
        d.star = runif(1, 0, b.star*.99)
        
        # debugging
        # b = 0.002419612
        # d = 0.001162986
        # b.star = 0.002419612
        # d.star = 0.001162986
        
        est = fbdR::est.bd.combined(tree1, frs, b = b, d = d, b.star = b.star, d.star = d.star, constrained = FALSE, crown = crown, rho = rho)
        
        convergence = est$convergence
        
        if(attempts == attempts.max)
          break
        
        attempts = attempts + 1
      }
      
      if(convergence != 0)
        next
      else{
        if(est$value > lnL){
          lnL = est$value
          tmp = est
        }
      }
      #if(debugging)
      #cat(est$par[1], "\t", est$par[2], "\t", est$par[3], "\t", est$par[4], "\t", est$value, "\n", file="debugging.txt", append=TRUE)
    }
    
    if(is.null(tmp)){
      m3.birth.extant = c(m3.birth.extant, NA)
      m3.death.extant = c(m3.death.extant, NA)
      m3.birth.ranges = c(m3.birth.ranges, NA)
      m3.death.ranges = c(m3.death.ranges, NA)
      m3.lk = c(m3.lk, NA)
      m3.convergence = c(m3.convergence, est$convergence)
    } else {
      est = tmp
      m3.birth.extant = c(m3.birth.extant, est$par[1])
      m3.death.extant = c(m3.death.extant, est$par[2])
      m3.birth.ranges = c(m3.birth.ranges, est$par[3])
      m3.death.ranges = c(m3.death.ranges, est$par[4])
      m3.lk = c(m3.lk, est$value)
      m3.convergence = c(m3.convergence, est$convergence)
    }
  }
  
  cat("finished M3...\n", file=file2, append=T)
  
  ## likelihood ratio tests
  
  LR1 = c()
  LR2 = c()
  LR3 = c()
  
  accept1 = c()
  accept2 = c()
  accept3 = c()
  accept4 = c()
  accept5 = c()
  accept6 = c()
  
  if(debugging)
    cat("Model testing...\n")
  
  for(i in 1:replicates){
    
    if(debugging)
      cat(i,"\n")
    
    lk1 = m1.lk[i]
    lk2 = m2.lk[i]
    lk3 = m3.lk[i]
    
    # model 1 (l = l*, m = m*) vs model 2 (l, l*, d), df = 0.5
    if(is.na(lk1) || is.na(lk2)){
      
      LR1 = c(LR1, NA)
      accept1 = c(accept1, NA)
      accept2 = c(accept2, NA)
      
    } else {
      
      LR = 2 * (lk2 - lk1)
      
      if(LR > qchisq(.95, df=df1))
        accept1 = c(accept1, 0) else
          accept1 = c(accept1, 1)
        
        if(LR > qchisq(.99, df=df1))
          accept2 = c(accept2, 0) else
            accept2 = c(accept2, 1)  
          
          LR1 = c(LR1, LR)
    }
    
    # model 1 (l = l*, m = m*) vs model 3 (l, l*, m, m*), df = 2
    if(is.na(lk1) || is.na(lk3)){
      
      LR2 = c(LR2, NA)
      accept3 = c(accept3, NA)
      accept4 = c(accept4, NA)
      
    } else {
      
      LR = 2 * (lk3 - lk1)
      
      if(LR > qchisq(.95, df=df2))
        accept3 = c(accept3, 0) else
          accept3 = c(accept3, 1)
        
        if(LR > qchisq(.99, df=df2))
          accept4 = c(accept4, 0) else
            accept4 = c(accept4, 1)        
          
          LR2 = c(LR2, LR)
    }
    
    # model 2 (l, l*, d) vs model 3 (l, l*, m, m*), df = 1.5
    if(is.na(lk2) || is.na(lk3)){
      
      LR3 = c(LR3, NA)
      accept5 = c(accept5, NA)
      accept6 = c(accept6, NA)
      
    } else {
      
      LR = 2 * (lk3 - lk2)
      
      # accept or reject the null model
      # reject the null if:
      if(LR > qchisq(.95, df=degrees.of.freedom))
        accept5 = c(accept5, 0) else
          accept5 = c(accept5, 1)
        
        if(LR > qchisq(.99, df=degrees.of.freedom))
          accept6 = c(accept6, 0) else
            accept6 = c(accept6, 1)    
          
          LR3 = c(LR3, LR)
    }
    #eol
  }
  
  cat("finished LR tests...\n", file=file2, append=T)
  
  ## log output
  
  if(debugging)
    cat("Printing output...\n")
  
  # print headers
  cat(# tree info
    "replicate", "\t",
    "ranges", "\t", # number of ranges
    "max_range", "\t", # max range age. Should = the origin time
    "tips", "\t", # this should be the same for all trees
    "origin", "\t",
    "crown", "\t",
    "rho", "\t",
    # model 1
    "m1_lambda", "\t",
    "m1_mu", "\t",
    "m1_lambda*", "\t",
    "m1_mu*", "\t",
    "m1_lk", "\t",
    "m1_convergence", "\t",
    # model 2
    "m2_lambda", "\t",
    "m2_mu", "\t",
    "m2_lambda*", "\t",
    "m2_mu*", "\t",
    "m2_lk", "\t",
    "m2_convergence", "\t",
    # model 3
    "m3_lambda", "\t",
    "m3_mu", "\t",
    "m3_lambda*", "\t",
    "m3_mu*", "\t",
    "m3_lk", "\t",
    "m3_convergence", "\t",
    # lk tests
    "LR_1v2", "\t",
    "accept_1v2_95", "\t",
    "accept_1v2_99", "\t",
    "LR_1v3", "\t",
    "accept_1v3_95", "\t",
    "accept_1v3_99", "\t",
    "LR_2v3", "\t",
    "accept_2v3_95", "\t",
    "accept_2v3_99", "\n",
    sep = "",
    file = file,
    append = F
  )
  
  for (i in 1:replicates){
    
    tree1<-trees[[i]]
    frs<-fossils[[i]]
    
    # collect useful info
    origin = max(FossilSim::symmetric.ages(tree1, root.edge = T)$start)
    t = geiger::drop.extinct(tree1)
    crown = max(FossilSim::symmetric.ages(t, root.edge = F)$start)
    tips = length(t$tip.label)
    
    cat(# tree info
      i, "\t",
      length(frs$sp), "\t", # number of ranges
      max(frs$start), "\t", # max range age. Should = the origin time
      tips, "\t", # this should be the same for all trees
      origin, "\t",
      crown, "\t",
      rho, "\t",
      # model 1
      m1.birth.extant[i], "\t",
      m1.death.extant[i], "\t",
      m1.birth.ranges[i], "\t",
      m1.death.ranges[i], "\t",
      m1.lk[i], "\t",
      m1.convergence[i], "\t",
      # model 2
      m2.birth.extant[i], "\t",
      m2.death.extant[i], "\t",
      m2.birth.ranges[i], "\t",
      m2.death.ranges[i], "\t",
      m2.lk[i], "\t",
      m2.convergence[i], "\t",
      # model 3
      m3.birth.extant[i], "\t",
      m3.death.extant[i], "\t",
      m3.birth.ranges[i], "\t",
      m3.death.ranges[i], "\t",
      m3.lk[i], "\t",
      m3.convergence[i], "\t",
      # lk tests
      LR1[i], "\t",
      accept1[i], "\t", 
      accept2[i], "\t",
      LR2[i], "\t",
      accept3[i], "\t",
      accept4[i], "\t",
      LR3[i], "\t",
      accept5[i], "\t",
      accept6[i], "\n",
      sep = "",
      file = file,
      append = T
    )
    
  }
  
  ## expected values
  
  lb = lambda.branching
  mb = mu.branching
  #ls = (1-f)*2*lambda.branching+f*lambda.branching+lambda.a
  #ms = (1-f)*lambda.branching+lambda.a+mu.branching
  rv = read.table(rv.file, header = T) # range values
  ls = rv$ls
  ms = rv$ms
  
  return(list(lb = lb, mb = mb, ls = ls, ms = ms,
              # model 1
              m1.birth.extant = m1.birth.extant,
              m1.death.extant = m1.death.extant,
              m1.birth.ranges = m1.birth.ranges,
              m1.death.ranges = m1.death.ranges,
              m1.lk = m1.lk,
              m1.convergence = m1.convergence,
              # model 2
              m2.birth.extant = m2.birth.extant,
              m2.death.extant = m2.death.extant,
              m2.birth.ranges = m2.birth.ranges,
              m2.death.ranges = m2.death.ranges,
              m2.lk = m2.lk,
              m2.convergence = m2.convergence,
              # model 3
              m3.birth.extant = m3.birth.extant,
              m3.death.extant = m3.death.extant,
              m3.birth.ranges = m3.birth.ranges,
              m3.death.ranges = m3.death.ranges,
              m3.lk = m3.lk,
              m3.convergence = m3.convergence,
              LR1 = LR1, LR2 = LR2, LR3 = LR3,
              accept1 = accept1, accept2 = accept2, accept3 = accept3,
              accept4 = accept4, accept5 = accept5, accept6 = accept6
  ))
  #eof
}

###

generate.histograms <- function(tree.file, fossil.file, dAIC.file, rho, label, nexus = T, log.file  = "", x.adjust = 0, h.just = 0.5){
  
  tmp<-read.table(dAIC.file,header = T)
  
  cat(label, "...\n", file = log.file, append = T)
  
  # read empirical data
  if(nexus)
    tr <- ape::read.nexus(file = tree.file)
  else
    tr <- ape::read.tree(file = tree.file)
  frs <- read.table(file = fossil.file, header=T)
  names(frs)[2]<-"sp"
  names(frs)[3]<-"start"
  names(frs)[4]<-"end"
  
  # contrained
  lnL = -10^100
  for(j in 1:recalculate.lk){
    attempts = 0
    convergence = 1
    while(convergence != 0){
      b = runif(1)
      b.star = runif(1, b, 2)
      nd = runif(1, 0, b*.99)
      est <- fbdR::est.bd.combined(tr,frs,b=b,b.star=b.star,nd=nd,crown=TRUE,rho=rho,constrained=TRUE)
      convergence = est$convergence
      if(attempts == 100)
        break
      attempts = attempts + 1
    }
    if(convergence != 0){
      next
    } else if(est$value > lnL){
      lnL = est$value
    }
    cat("j ", j, " lnL null = ", lnL, "\n", sep="", file = log.file, append = T)
  }
  if(convergence != 0)
    next
  AIC.null = -(2 * lnL) + (2 * 2.5) # should this be 3 or 2.5?
  
  # unconstrained
  lnL = -10^100
  for(j in 1:recalculate.lk){
    attempts = 0
    convergence = 1
    while(convergence != 0){
      b = runif(1)
      d = runif(1, 0, b*.99)
      b.star = runif(1, b, 2)
      d.star = runif(1, 0, b.star*.99)
      est <- fbdR::est.bd.combined(tr,frs,b=b,d=d,b.star=b.star,d.star=d.star,crown=TRUE,rho=rho,constrained=FALSE)
      convergence = est$convergence
      if(attempts == 100)
        break
      attempts = attempts + 1
    }
    if(convergence != 0){
      next
    } else if(est$value > lnL){
      lnL = est$value
    }
    cat("j ", j, " lnL alt = ", lnL, "\n", sep="", file = log.file, append = T)
  }
  if(convergence != 0)
    next
  AIC.alt = -(2 * lnL) + (2 * 4)
  
  dAIC.empirical = AIC.null - AIC.alt
  
  # simulated data
  dAIC.dist <- tmp$dAIC
  #dd = dAIC.dist
  # remove extremes
  qs = quantile(dAIC.dist,probs=c(.01,.99)) # alternatives: 0.025, 0.975; 0.05, 0.95; 
  dd = dAIC.dist[which(dAIC.dist > qs[[1]])]
  dd = dd[which(dd < qs[[2]])]
  
  dat<-data.frame(d=dd,id=1)
  
  xe = (min(dd,dAIC.empirical) + max(dd,dAIC.empirical)) / 2 
  
  cat("Informative dAICs =", length(dd), "\n", file = log.file, append = T)
  
  p = ggplot(dat,aes(x=d)) +
    geom_histogram(fill = "dodgerblue3",bins=10,alpha=.75) +
    geom_vline(xintercept = dAIC.empirical, colour = "tomato") +
    labs(title = label, x = "dAIC") +
    theme(legend.position="none")
  
  build = ggplot_build(p)
  y.range = build$panel$ranges[[1]]$y.range
  ye = (y.range[1] + y.range[2]) * .25  
  
  p = p + geom_label(x=dAIC.empirical+x.adjust,y=ye,aes(label=paste("dAIC =",round(dAIC.empirical,2)),size=2.5,hjust=h.just),fill="tomato",col="white")
  
  return(p)
  #eof
}

###

generate.histograms.LR <- function(tree.file, fossil.file, LR.file, rho, label, nexus = T, log.file  = "", x.adjust = 0, h.just = 0.5){
  
  tmp<-read.table(LR.file,header = T)
  
  cat(label, "...\n", file = log.file, append = T)
  
  # read empirical data
  if(nexus)
    tr <- ape::read.nexus(file = tree.file)
  else
    tr <- ape::read.tree(file = tree.file)
  frs <- read.table(file = fossil.file, header=T)
  names(frs)[2]<-"sp"
  names(frs)[3]<-"start"
  names(frs)[4]<-"end"
  
  # contrained
  lnL = -10^100
  for(j in 1:recalculate.lk){
    attempts = 0
    convergence = 1
    while(convergence != 0){
      b = runif(1)
      b.star = runif(1, b, 2)
      nd = runif(1, 0, b*.99)
      est <- fbdR::est.bd.combined(tr,frs,b=b,b.star=b.star,nd=nd,crown=TRUE,rho=rho,constrained=TRUE)
      convergence = est$convergence
      if(attempts == 100)
        break
      attempts = attempts + 1
    }
    if(convergence != 0){
      next
    }
    if(est$value > lnL){
      lnL = est$value
    }
    cat("j ", j, " lnL null = ", lnL, "\n", sep="", file = log.file, append = T)
  }
  if(convergence != 0)
    next
  lk.null = lnL
  
  # unconstrained
  lnL = -10^100
  for(j in 1:recalculate.lk){
    attempts = 0
    convergence = 1
    while(convergence != 0){
      b = runif(1)
      d = runif(1, 0, b*.99)
      b.star = runif(1, b, 2)
      d.star = runif(1, 0, b.star*.99)
      est <- fbdR::est.bd.combined(tr,frs,b=b,d=d,b.star=b.star,d.star=d.star,crown=TRUE,rho=rho,constrained=FALSE)
      convergence = est$convergence
      if(attempts == 100)
        break
      attempts = attempts + 1
    }
    if(convergence != 0){
      next
    } else if(est$value > lnL){
      lnL = est$value
    }
    cat("j ", j, " lnL alt = ", lnL, "\n", sep="", file = log.file, append = T)
  }
  if(convergence != 0)
    next
  lk.alt = lnL
  
  # calculate lk ratio
  LR.empirical = 2 * (lk.alt - lk.null)
  
  # simulated data
  LR.dist <- tmp$LR
  #dd = LR.dist
  # remove extremes
  qs = quantile(LR.dist,probs=c(.01,.99)) #.05,.95
  dd = LR.dist[which(LR.dist > qs[[1]])]
  dd = dd[which(dd < qs[[2]])]
  
  dat<-data.frame(d=dd,id=1)
  
  xe = (min(dd,LR.empirical,qchisq(.95, df=degrees.of.freedom)) + max(dd,LR.empirical,qchisq(.99, df=degrees.of.freedom)))  / 2
  
  cat("Informative LRs =", length(dd), "\n", file = log.file, append = T)
  
  p = ggplot(dat,aes(x=d)) +
    geom_histogram(fill = "dodgerblue3",bins=10,alpha=.75) +
    geom_vline(xintercept = LR.empirical, colour = "tomato") +
    geom_vline(xintercept = qchisq(.95, df=degrees.of.freedom), colour = "black", lty = 2) +
    geom_vline(xintercept = qchisq(.99, df=degrees.of.freedom), colour = "black", lty = 2) +
    labs(title = label, x = "LR") +
    theme(legend.position="none")
  
  build = ggplot_build(p)
  y.range = build$panel$ranges[[1]]$y.range
  ye = (y.range[1] + y.range[2]) * .25  
  
  ychi1 = (y.range[1] + y.range[2]) * .5
  ychi2 = (y.range[1] + y.range[2]) * .75
  
  lb1 = "         "
  
  p = p + geom_label(x=LR.empirical+x.adjust,y=ye,aes(label=paste("LR =",round(LR.empirical,2)),size=2.5,hjust=h.just),fill="tomato",col="white") +
      # chi squared .95
      geom_label(x=qchisq(.95, df=degrees.of.freedom),y=ychi1,aes(label=lb1,size=2.5, parse = TRUE),fill="darkgrey",col="white") +
      annotate("text", x = qchisq(.95, df=degrees.of.freedom), y = ychi1, label = "x^{2}~.95", parse = TRUE, col = "white") +
      # chi squared .99
      geom_label(x = (qchisq(.99, df=degrees.of.freedom)),y = ychi2,aes(label=lb1,size=2.5, parse = TRUE, hjust = 0.7),fill="darkgrey",col="white") +
      annotate("text", x = (qchisq(.99, df=degrees.of.freedom)), y = ychi2, label = "x^{2}~.99", parse = TRUE, col = "white", hjust = 0.7)
      # geom_label(x=qchisq(.99, df=1),y=ychi,aes(label=".99",size=2.5),fill="darkgrey",col="white")
  
  return(p)
  #eof
}

### 

##### # table 1 - estimates & likelihood ratio test

print.headers<-function(file = "lk_ratio_test_results_simulations.txt"){
  
  cat("lambda", "\t",
      "mu", "\t",
      "lambda*", "\t",
      "mu*", "\t",
      "d", "\t",
      "d*", "\t",
      # model 1
      "m1_l","\t",
      "m1_m","\t",
      "m1_l*","\t",
      "m1_m*","\t",
      "m1_d","\t",
      "m1_d*","\t",
      # model 2
      "m2_l","\t",
      "m2_m","\t",
      "m2_l*","\t",
      "m2_m*","\t",
      "m2_d","\t",
      "m2_d*","\t",
      # model 3
      "m3_l","\t",
      "m3_m","\t",
      "m3_l*","\t",
      "m3_m*","\t",
      "m3_d","\t",
      "m3_d*","\t",
      # lk ratio
      "m1_v_m2_95","\t",
      "m1_v_m2_99","\t",
      "m1_v_m3_95","\t",
      "m1_v_m3_99","\t",
      "m2_v_m3_95","\t",
      "m2_v_m3_99","\n", # this is the original test
      file = file, sep = "",
      append = F
  )
  #eof
}

print.results<-function(p, file = "lk_ratio_test_results_simulations.txt"){
  
  cat(round(mean(p$lb, na.rm = T),3), "\t",
      round(mean(p$mb, na.rm = T),3), "\t",
      round(mean(p$ls, na.rm = T),3), "\t",
      round(mean(p$ms, na.rm = T),3), "\t",
      round(mean(p$lb - p$mb, na.rm = T),3), "\t",
      round(mean(p$ls - p$ms, na.rm = T),3), "\t",
      # model 1
      round(mean(p$m1.birth.extant, na.rm = T),3),"\t",
      round(mean(p$m1.death.extant, na.rm = T),3),"\t",
      round(mean(p$m1.birth.ranges, na.rm = T),3),"\t",
      round(mean(p$m1.death.ranges, na.rm = T),3),"\t",
      round(mean(p$m1.birth.extant - p$m1.death.extant, na.rm = T),3), "\t",
      round(mean(p$m1.birth.ranges - p$m1.death.ranges, na.rm = T),3), "\t",
      # model 2
      round(mean(p$m2.birth.extant, na.rm = T),3),"\t",
      round(mean(p$m2.death.extant, na.rm = T),3),"\t",
      round(mean(p$m2.birth.ranges, na.rm = T),3),"\t",
      round(mean(p$m2.death.ranges, na.rm = T),3),"\t",
      round(mean(p$m2.birth.extant - p$m2.death.extant, na.rm = T),3), "\t",
      round(mean(p$m2.birth.ranges - p$m2.death.ranges, na.rm = T),3), "\t",
      # model 3
      round(mean(p$m3.birth.extant, na.rm = T),3),"\t",
      round(mean(p$m3.death.extant, na.rm = T),3),"\t",
      round(mean(p$m3.birth.ranges, na.rm = T),3),"\t",
      round(mean(p$m3.death.ranges, na.rm = T),3),"\t",
      round(mean(p$m3.birth.extant - p$m3.death.extant, na.rm = T),3), "\t",
      round(mean(p$m3.birth.ranges - p$m3.death.ranges, na.rm = T),3), "\t",
      # lk ratio
      round(sum(p$accept1, na.rm = T)/length(na.omit(p$accept1)),3),"\t",
      round(sum(p$accept2, na.rm = T)/length(na.omit(p$accept2)),3),"\t",
      round(sum(p$accept3, na.rm = T)/length(na.omit(p$accept3)),3),"\t",
      round(sum(p$accept4, na.rm = T)/length(na.omit(p$accept4)),3),"\t",
      round(sum(p$accept5, na.rm = T)/length(na.omit(p$accept5)),3),"\t",
      round(sum(p$accept6, na.rm = T)/length(na.omit(p$accept6)),3),"\n",
      file = file, sep = "",
      append = T
  )
  
  # eof
}


#END
